'
'Matrix library,using F32 for floating point math.
'
'

Con
OBJ
F32 : "f32"
Var
  long Matlist[20]
  long mattrack[20]
  long varspace[200]
  long edgespace
  long Sedgespace

  long ANS[50]
  long seed
  long endmat
PUB clear
  'clears the variable space, call first
  longfill(@varspace, 0, 200)
  F32.start
  edgespace := 0
  long[@Matlist][20] := @ANS
  seed := 15

PUB MCreate(row,col) :Retval | ptr, i, mnum
  'create a mxn matrix, return its id to be used in other functions
  ptr := @varspace[edgespace]
  mnum := NextAvail
  
  byte[ptr][0] := col*row + 1 
  byte[ptr][1] := row
  byte[ptr][2] := col
  byte[ptr][3] := mnum
  
  long[@Matlist][mnum] := ptr
  
  Retval := mnum
  

  edgespace += col*row+1 

PUB NextAvail :retval | i
  'find an available space in the memory pointer array
  repeat i from 1 to 20
    if long[@Matlist][i] == 0
      retval := i
      quit 
PUB retmatlist(num): retva
  retva := long[@Matlist][num]
'PUB MLookup(Num) :retval
'  retval := long[@Matlist][Num]       
PUB MDelete(MAT) | dsize, front, count, end, i, addr
  'delete a given matrix, ID is no longer valid, and space is shifted up
  addr := long[@Matlist][MAT] 
  dsize := byte[addr][0]
  front := addr + dsize*4 '6772 to 6789
  end :=   @varspace[0]+ edgespace*4'ans 6750

  count :=  end - front
  long[@Matlist][MAT] := 0  
  repeat i from MAT to 20
    if long[@Matlist][i] <> 0 and i <> 20
     long[@Matlist][i] -= dsize*4

  bytemove(addr,front, count)
  edgespace -= dsize
  long[@varspace][edgespace] := 0

PUB vspace : retv
  retv := @varspace[0]  
PUB WVal(MAT, Rloc,Cloc, Value) | vloc
  'write a integer value into matrix space, not usually used as everything is floating point
  MAT := long[@Matlist][MAT]
  vloc :=  (Rloc-1)*byte[MAT][2] + Cloc
  long[MAT][vloc] := Value
PUB WFVal(MAT, Rloc,Cloc, Value) | vloc
  'write a floating point number into the matrix,
  '  WFVal(matrix,row,col, values)    Value is 32bit floating point number
  MAT := long[@Matlist][MAT]  
  vloc :=  (Rloc-1)*byte[MAT][2] + Cloc
  long[MAT][vloc] := Value
PUB WFValInt(MAT, Rloc,Cloc, Value) | vloc
  'write a floating point number into the matrix,
  '  WFVal(matrix,row,col, values)    Value is an integer
  MAT := long[@Matlist][MAT]
  
  vloc :=  (Rloc-1)*byte[MAT][2] + Cloc
  long[MAT][vloc] := F32.FFloat(Value)  
PUB RFVal(MAT, Rloc,Cloc) :Retval | vloc
  'read a value from the matrix, Returned is a floating point number
  'RFVal(Matrix,row,col)
  
  MAT := long[@Matlist][MAT]
  vloc :=(Rloc-1)*byte[MAT][2] + Cloc
  Retval := long[MAT][vloc]

PUB MMult(A,B): retans |i,j,k,rowa,cola,colb,rowb, temp, temp2
  'Multiply a matrix AXB and store it in the ANS space
  'To keep result, use MCopy to copy in another matrix
  rowa := MRow(A)
  cola := MCol(A)
  rowb := MRow(B)
  colb := MCol(B) 
  temp := constant(0.0)
  ANSSMClear
  retans := ANSMCreate(rowa,colb)
  
  if cola == rowb
    repeat i from 1 to rowa
      repeat j from 1 to colb
        temp := constant(0.0)
        repeat k from 1 to rowb
         temp2 := F32.Fmul(RFVal(A,i,k),RFVal(B,k,j))
         temp := F32.Fadd(temp,temp2)
        WFval(retans,i,j,temp)
PUB ScalarMMult(A,Scalar)| i,j, temp, row,col 
  'multiplies the current matrix by a scalar, saves in same spot
  'scalar is a floating point number
  row := MRow(A)
  col := MCol(A)
  repeat i from 1 to row
    repeat j from 1 to col
      temp :=  F32.FMul(RFVal(A,i,j),scalar)
      WFVal(A,i,j,temp)
PUB MTrace(A):retval | i,j, temp, row,col 
  'multiplies the current matrix by a scalar, saves in same spot
  'scalar is a floating point number
  row := MRow(A)
  col := MCol(A)
  retval := constant(0.0)
  repeat i from 1 to row
      retval :=  F32.FADD(RFVal(A,i,i),retval)
    
PUB MSub(A,B) : retans | i,j, temp, row,col
  'subtract matrix A from B, stores in the ANS space
  row := MRow(A)
  col := MCol(A)
  ANSSMClear
  retans := ANSMCreate(row,col)

  repeat i from 1 to row
    repeat j from 1 to col
      temp := F32.FSub(RFVal(A,i,j), RFVal(B,i,j))
      WFVal(retans,i,j,temp)
PUB MAdd(A,B) : retans | i,j, temp, row,col
  'ADD matrix A from B, stores in ANS space
  row := MRow(A)
  col := MCol(A)
  ANSSMClear
  retans := ANSMCreate(row,col)

  repeat i from 1 to row
    repeat j from 1 to col
      temp := F32.FAdd(RFVal(A,i,j), RFVal(B,i,j))
      WFVal(retans,i,j,temp)      


  
  
PUB Skew3(A) : retans | i,j
  ' puts a vector into a skew 3,3 matrix
  ANSSMClear
  retans := ANSMCreate(3,3)

  WFVal(retans,1,2,F32.FNeg(RFVal(A,3,1)))
  WFVal(retans,1,3,RFVal(A,2,1))
  WFVal(retans,2,3, F32.FNeg(RFVal(A,1,1)))

  WFVal(retans,2,1,RFVal(A,3,1) )
  WFVal(retans,3,1,F32.FNeg(RFVal(A,2,1))) 
  WFVal(retans,3,2, RFVal(A,1,1) )     
  
PUB RFValloc(MAT, Rloc,Cloc) :Retval | vloc
  vloc :=(Rloc-1)*byte[MAT][2] + Cloc
  Retval := MAT+vloc
  
PUB Transpose(MAT)  | sptr,row, col, i,j, temp, temp2, tempMat
  'Transposes a matrix, Matrix is copied into the origonal, to under simply transpose back.
  row := MRow(Mat)
  col := MCol(Mat)

  ANSSMCLEAR
  tempMat := ANSMCreate(row,col)
  MCopy(MAT,tempMat)

  byte[MAT][1] := col
  byte[MAT][2] := row
   repeat i from 1 to row
     repeat j from 1 to col    
       WFVal(MAT,j,i, RFVal(tempMat,i,j))     
  
  
  
PUB MRow(MAT) : retval
  'return the row size of a matrix
  MAT := long[@Matlist][MAT]
    retval := byte[MAT][1]
PUB MCol(MAT) : retval
  'return the column size of a matrix
  MAT := long[@Matlist][MAT]
    retval := byte[MAT][2]


PUB MCopy(A,B)
  'Copu matrix A, into Matrix B
  A := long[@Matlist][A]
  B := long[@Matlist][B]
'copy mat A into B
  if byte[A][1] == byte[B][1] and byte[A][2] == byte[B][2]  
    longmove(B,A,byte[B][0])  
PUB espace : rets
  rets := edgespace
PUB Mrandom(row,col) : retran | i,j
  'Generate a random mxn matrix
  ANSSMClear
  retran := ANSMCreate(row,col)
  repeat i from 1 to row
    repeat j from 1 to col
      WFVal(retran,i,j,F32.FDiv(F32.FFloat(?seed), F32.FFloat(?seed)))'
PUB LUDecomp(MAT) : RetANS  | row, col, i, j, k,L,U , temp,t1,t2
  'Computes Lower and Upper matrices using  LU decompisition
  'stores the matrices in a single nxn matrix, where the upper and low portions
  'both exist, the diagonal 1's are assumed on lower
  row := MRow(MAT)
  col := MCol(MAT)
  ANSSMCLEAR
  RetANS := ANSMCreate(row,col)
  MCopy(MAT, RetANS)
  repeat k from 1 to col-1
    repeat i from k+1 to col
      t1 := F32.FDiv(RFVal(RetANS,i,k),RFVal(RetANS,k,k))
      WFVal(RetANS,i,k,t1)
      repeat j from k+1 to col
        t2 := F32.FMul(t1,RFVal(RetANS,k,j))
        WFVal(RetANS,i,j, F32.FSub(RFVal(RetANS,i,j), t2))
PUB MINV(A):retans |i,j,row,col, temp, r1,r2,Ainv,addr
  'invert a matrix, using LUSolve function to fill in each row. return the answer in ANS space
  row := MRow(A)
  col := MCol(A)

  r1 := MCreate(row,1)
  r2 := MCreate(row,1)
  Ainv := MCreate(row,col)

  repeat j from 1 to col
    addr := long[@Matlist][r1]
    longfill(addr+4, 0, row )
    WFValInt(r1,j,1,1)
    temp :=  LUSolve(A,r1,1)
    repeat i from 1 to row
      WFVal(Ainv,i,j, RFVal(temp,i,1))   
  
  ANSSMClear
  temp := ANSMCreate(row,col)
  MCopy(Ainv,temp)
  retans := temp
  MDelete(r1)
  MDelete(r2)
  MDelete(Ainv)
 ' retans := r1  
PUB LUSolve(A,B,piv):retans | LUMAT,L,U,i,j,row,col,temp,D, dex,tempa,tempb
  'A is the mxm matrix, B is the mx1 RHS vector, solves the system [A][X] = [B]
  'LUSolve(A,B,pivoting)   pivoting = 1 enabled, 0 disabled
  'disable pivoting only when trace is not approx 0 and less accuracy is fine
  row := MRow(A)
  col := MCol(A)

  'SMClear   
  ANSSMClear
  LUMAT := MCreate(row,col)
  tempb := MCreate(row,1)
  D := MCreate(row,1)
  dex := MCreate(row,1)
  'pivot scheme
  MCopy(A,LUMat)
  MCopy(B,tempb)
  if piv == 1
    repeat i from 1 to col
     WVal(dex,i,1,i)
    
    repeat i from 1 to col
     retans := MPivot(LUMat,tempb,i,dex)
 ' 
  temp := LUDecomp(LUMat)
  MCopy(temp,LUMAT)
 '
  temp := LUForwardSub(LUMAT,tempb)
  MCopy(temp,D)
  retans := LUBackwardSub(LUMAT,D)
  'retans := LUMAT
  MDelete(LUMAT)
  MDelete(tempb)
  MDelete(D)
  MDelete(dex)
PUB MPivot(A,B,k, dex):retans | i,j,t1, temp, index, bi, big,row,col
  'pivot rows by changing their location by copying the values, indexing will be much faster
  row := MRow(A)
  col := MCol(A)

  bi := k
  big := F32.FAbs(RFVal(A,k,k))



  repeat i from k to col
    temp := F32.FAbs(RFVal(A,i,k))
    t1 := F32.FCmp(temp,big)
    if t1 == 1
      bi := i
      big := temp

      

  if bi <> k

    repeat j from 1 to col
      temp := RFVal(A,bi,j)
      WFVal(A,bi,j, RFVal(A,k,j))
      WFVal(A,k,j, temp)

    temp := RFVal(B,bi,1)
    WFVal(B,bi,1, RFVal(B,k,1))
    WFVal(B,k,1, temp)

    temp := RFVal(dex,bi,1)
    WVal(dex,bi,1, RFVal(dex,k,1))
    WVal(dex,k,1, temp)
      
        

      
  


PUB LUBackwardSub(A, D): retans |i,j,col,row,t1,t2
  'backward solve the matrix
  row := MRow(A)
  col := MCol(A)

  ANSSMClear
  retans := ANSMCreate(row,1)

  'write first val
  t1 := F32.FDiv(RFVal(D,col,1), RFVal(A,col,col))
  WFVal(retans,col,1, t1)

  repeat i from col -1 to 1
    t1 := F32.FFLoat(0)
    repeat j from i+1 to col
      t2 := F32.FMul(RFVal(A,i,j),RFVal(retans,j,1))
      t1 := F32.Fadd(t1,t2)
    t2 := F32.FSub(RFVal(D,i,1),t1)
    t1 := F32.FDiv(t2,RFVal(A,i,i))
          
    WFVal(retans,i,1,t1)      
PUB LUForwardSub(A, B): retans |i,j,col,row,t1,t2
  'forward solve a matrix
  row := MRow(A)
  col := MCol(A)

  ANSSMClear
  retans := ANSMCreate(row,1)
  WFVal(retans,1,1, RFVal(B,1,1))  
  repeat i from 2 to row
    t1 := F32.FFloat(0)
    repeat j from 1 to i-1
      t2 := F32.FMul(RFVal(A,i,j),RFVal(retans,j,1))
      t1 := F32.Fadd(t1,t2)
    t2 := F32.FSub(RFVal(B,i,1),t1)
    WFVal(retans,i,1,t2)
    
pub addfp(A,B) : retans
  'add a single floating point value A to floating point value B
   retans := F32.Fadd(A,B)
pub subfp(A,B) : retans
  'add a single floating point value A to floating point value B
   retans := F32.FSub(A,B)
pri ANSMCreate(row,col) :Retval | ptr, i

  ptr := @ANS[0]

  byte[ptr][0] := col*row+1 
  byte[ptr][1] := row
  byte[ptr][2] := col
  byte[ptr][3] := 20
  long[@Matlist][20] := ptr 
  'Sedgespace += col*row+1 
  Retval := 20
pri ANSSMClear
  longfill(@ANS, 0, 50)
   long[@Matlist][20] := 0